@@ -1,5 +1,3 @@
-import six
-
 from gssapi.raw import sec_contexts as rsec_contexts
 from gssapi.raw import message as rmessage
 from gssapi.raw import named_tuples as tuples
@@ -11,8 +9,8 @@ from gssapi.names import Name
 from gssapi.creds import Credentials
 
 
-@six.add_metaclass(_utils.CheckLastError)
-class SecurityContext(rsec_contexts.SecurityContext):
+class SecurityContext(rsec_contexts.SecurityContext,
+                      metaclass=_utils.CheckLastError):
     """A GSSAPI Security Context
 
     This class represents a GSSAPI security context that may be used
@@ -22,7 +20,8 @@ class SecurityContext(rsec_contexts.SecurityContext):
     :class:`~gssapi.raw.sec_contexts.SecurityContext` class,
     and thus may used with both low-level and high-level API methods.
 
-    This class may be pickled an unpickled.
+    This class may be pickled and unpickled (the attached delegated
+    credentials object will not be preserved, however).
     """
 
     def __new__(cls, base=None, token=None,
@@ -106,14 +105,25 @@ class SecurityContext(rsec_contexts.SecurityContext):
             self._channel_bindings = channel_bindings
             self._creds = creds
 
-            self.delegated_creds = None
+            self._delegated_creds = None
 
         else:
             # we already have a context in progress, just inspect it
-            if self.locally_initiated:
-                self.usage = 'initiate'
-            else:
-                self.usage = 'accept'
+            # NB(directxman12): MIT krb5 refuses to inquire about a context
+            # if it's partially established, so we have to check here
+
+            try:
+                if self.locally_initiated:
+                    self.usage = 'initiate'
+                else:
+                    self.usage = 'accept'
+            except excs.MissingContextError:
+                msg = ("Cannot extract usage from a partially completed "
+                       "context")
+                raise excs.UnknownUsageError(msg, obj="security context")
+
+        # This is to work around an MIT krb5 bug (see the `complete` property)
+        self._complete = None
 
     # NB(directxman12): DO NOT ADD AN __del__ TO THIS CLASS -- it screws up
     #                   the garbage collector if _last_tb is still defined
@@ -121,7 +131,7 @@ class SecurityContext(rsec_contexts.SecurityContext):
     # TODO(directxman12): implement flag properties
 
     def get_signature(self, message):
-        """Calculate the signature for a message
+        """Calculate the signature for a message.
 
         This method calculates the signature (called a MIC) for
         the given message, which may be then used with
@@ -134,16 +144,21 @@ class SecurityContext(rsec_contexts.SecurityContext):
 
         Returns:
             bytes: the message signature
+
+        Raises:
+            ExpiredContextError
+            MissingContextError
+            BadQoPError
         """
 
         # TODO(directxman12): check flags?
         return rmessage.get_mic(self, message)
 
     def verify_signature(self, message, mic):
-        """Verify the signature for a message
+        """Verify the signature for a message.
 
         This method verifies that a signature (generated by
-        :meth:`get_signature` is valid for the given method.
+        :meth:`get_signature` is valid for the given message.
 
         If the signature is valid, the method will return.
         Otherwise, it will raise an error.
@@ -154,6 +169,13 @@ class SecurityContext(rsec_contexts.SecurityContext):
 
         Raises:
             BadMICError: the signature was not valid
+            InvalidTokenError
+            DuplicateTokenError
+            ExpiredTokenError
+            TokenTooLateError
+            TokenTooEarlyError
+            ExpiredContextError
+            MissingContextError
         """
 
         return rmessage.verify_mic(self, message, mic)
@@ -161,8 +183,8 @@ class SecurityContext(rsec_contexts.SecurityContext):
     def wrap(self, message, encrypt):
         """Wrap a message, optionally with encryption
 
-        This method generates a signature and uses it to
-        wrap the message, optionally encrypting it.
+        This wraps a message, signing it and optionally
+        encrypting it.
 
         Args:
             message (bytes): the message to wrap
@@ -171,12 +193,17 @@ class SecurityContext(rsec_contexts.SecurityContext):
         Returns:
             WrapResult: the wrapped message and details about it
                 (e.g. whether encryption was used succesfully)
+
+        Raises:
+            ExpiredContextError
+            MissingContextError
+            BadQoPError
         """
 
         return rmessage.wrap(self, message, encrypt)
 
     def unwrap(self, message):
-        """Unwrap a wrapped message
+        """Unwrap a wrapped message.
 
         This method unwraps/unencrypts a wrapped message,
         verifying the signature along the way.
@@ -187,12 +214,22 @@ class SecurityContext(rsec_contexts.SecurityContext):
         Returns:
             UnwrapResult: the unwrapped message and details about it
                 (e.g. wheter encryption was used)
+
+        Raises:
+            InvalidTokenError
+            BadMICError
+            DuplicateTokenError
+            ExpiredTokenError
+            TokenTooLateError
+            TokenTooEarlyError
+            ExpiredContextError
+            MissingContextError
         """
 
         return rmessage.unwrap(self, message)
 
     def encrypt(self, message):
-        """Encrypt a message
+        """Encrypt a message.
 
         This method wraps and encrypts a message, similarly to
         :meth:`wrap`.  The difference is that encryption is always
@@ -208,6 +245,9 @@ class SecurityContext(rsec_contexts.SecurityContext):
 
         Raises:
             EncryptionNotUsed: the encryption could not be used
+            ExpiredContextError
+            MissingContextError
+            BadQoPError
         """
 
         res = self.wrap(message, encrypt=True)
@@ -218,12 +258,13 @@ class SecurityContext(rsec_contexts.SecurityContext):
         return res.message
 
     def decrypt(self, message):
-        """Decrypt a message
+        """Decrypt a message.
 
         This method decrypts and unwraps a message, verifying the signature
         along the way, similarly to :meth:`unwrap`.  The difference is that
-        this method will raise an exception if encryption was by the context
-        and not used, and simply returns the decrypted message directly.
+        this method will raise an exception if encryption was established
+        by the context and not used, and simply returns the decrypted
+        message directly.
 
         Args:
             message (bytes): the encrypted message
@@ -233,6 +274,14 @@ class SecurityContext(rsec_contexts.SecurityContext):
 
         Raises:
             EncryptionNotUsed: encryption was expected, but not used
+            InvalidTokenError
+            BadMICError
+            DuplicateTokenError
+            ExpiredTokenError
+            TokenTooLateError
+            TokenTooEarlyError
+            ExpiredContextError
+            MissingContextError
         """
 
         res = self.unwrap(message)
@@ -248,10 +297,10 @@ class SecurityContext(rsec_contexts.SecurityContext):
 
     def get_wrap_size_limit(self, desired_output_size,
                             encrypted=True):
-        """Get the maximum message size for a given wrapped message size
+        """Calculate the maximum message size for a given wrapped message size.
 
         This method calculates the maximum input message size for a given
-        wrapped/encrypted message size.
+        maximum wrapped/encrypted message size.
 
         Args:
             desired_output_size (int): the maximum output message size
@@ -260,31 +309,48 @@ class SecurityContext(rsec_contexts.SecurityContext):
 
         Returns:
             int: the maximum input message size
+
+        Raises:
+            MissingContextError
+            ExpiredContextError
+            BadQoPError
         """
 
         return rmessage.wrap_size_limit(self, desired_output_size,
                                         encrypted)
 
     def process_token(self, token):
-        """Process an output token asynchronously
+        """Process an output token asynchronously.
 
         This method processes an output token even when the security context
         was not expecting it.
 
+        Warning:
+            This method is deprecated.
+
         Args:
             token (bytes): the token to process
+
+        Raises:
+            InvalidTokenError
+            MissingContextError
         """
 
         rsec_contexts.process_context_token(self, token)
 
     def export(self):
-        """Export a security context
+        """Export a security context.
 
         This method exports a security context, allowing it to be passed
         between processes.
 
         Returns:
             bytes: the exported security context
+
+        Raises:
+            ExpiredContextError
+            MissingContextError
+            OperationUnavailableError
         """
 
         return rsec_contexts.export_sec_context(self)
@@ -306,7 +372,7 @@ class SecurityContext(rsec_contexts.SecurityContext):
             initiator_name (bool): get the initiator name for this context
             target_name (bool): get the target name for this context
             lifetime (bool): get the remaining lifetime for this context
-            mech (bool): get the mechanism used by this context
+            mech (bool): get the :class:`MechType` used by this context
             flags (bool): get the flags set on this context
             locally_init (bool): get whether this context was locally initiated
             complete (bool): get whether negotiation on this context has
@@ -315,6 +381,9 @@ class SecurityContext(rsec_contexts.SecurityContext):
         Returns:
             InquireContextResult: the results of the inquiry, with unused
                 fields set to None
+
+        Raises:
+            MissingContextError
         """
         if not kwargs:
             default_val = True
@@ -345,38 +414,62 @@ class SecurityContext(rsec_contexts.SecurityContext):
 
     @property
     def lifetime(self):
-        """Get the amount of time for which the context remains valid"""
+        """The amount of time for which this context remains valid"""
         return rsec_contexts.context_time(self)
 
+    @property
+    def delegated_creds(self):
+        """The credentials delegated from the initiator to the acceptor
+
+        .. warning::
+
+            This value will not be preserved across picklings.  These should
+            be separately exported and transfered.
+
+        """
+        return self._delegated_creds
+
     initiator_name = _utils.inquire_property(
-        'initiator_name', 'Get the Name of the initiator of this context')
+        'initiator_name', 'The :class:`Name` of the initiator of this context')
     target_name = _utils.inquire_property(
-        'target_name', 'Get the Name of the target of this context')
+        'target_name', 'The :class:`Name` of the target of this context')
     mech = _utils.inquire_property(
-        'mech', 'Get the mechanism in use by this context')
+        'mech', 'The mechanism (:class:`MechType`) in use by this context')
     actual_flags = _utils.inquire_property(
-        'flags', 'Get the flags set on this context')
+        'flags', 'The flags set on this context')
     locally_initiated = _utils.inquire_property(
-        'locally_init', 'Get whether this context was locally intiated')
+        'locally_init', 'Whether this context was locally intiated')
 
     @property
     @_utils.check_last_err
     def complete(self):
-        """Get whether negotiation for this context has been completed"""
+        """Whether negotiation for this context has been completed"""
+        # NB(directxman12): MIT krb5 has a bug where it refuses to
+        #                   inquire about partially completed contexts,
+        #                   so we can't just use `self._inquire` generally
         if self._started:
-            return self._inquire(complete=True).complete
+            if self._complete is None:
+                try:
+                    res = self._inquire(complete=True).complete
+                except excs.MissingContextError:
+                    return False
+                else:
+                    self._complete = res
+
+            return self._complete
         else:
             return False
 
     @_utils.catch_and_return_token
     def step(self, token=None):
-        """Perform a negotation step
+        """Perform a negotation step.
 
         This method performs a negotiation step based on the usage type
-        of this context.  If `__DEFER_STEP_ERRORS__` is set to true,
-        this method will return a token, even when exceptions would be
-        thrown.  The generated exception will be thrown on the next
+        of this context.  If `__DEFER_STEP_ERRORS__` is set to True on
+        the class, this method will return a token, even when exceptions
+        would be thrown.  The generated exception will be thrown on the next
         method call or property lookup on the context.
+        **This is the default behavior.**
 
         This method should be used in a while loop, as such:
 
@@ -390,11 +483,34 @@ class SecurityContext(rsec_contexts.SecurityContext):
            except GSSError as e:
                 handle_the_issue()
 
+        .. tip::
+
+            Disabling `__DEFER_STEP_ERRORS__` is rarely necessary.
+            When this method is used in a loop (as above),
+            `__DEFER_STEP_ERRORS__` will ensure that you always
+            send an error token when it's available,
+            keeping the other end of the security context updated
+            with the status of the negotiation.
+
         Args:
             token (bytes): the input token from the other participant's step
 
         Returns:
             bytes: the output token to send to the other participant
+
+        Raises:
+            InvalidTokenError
+            InvalidCredentialsError
+            MissingCredentialsError
+            ExpiredCredentialsError
+            BadChannelBindingsError
+            BadMICError
+            ExpiredTokenError: (initiate only)
+            DuplicateTokenError
+            MissingContextError
+            BadNameTypeError: (initiate only)
+            BadNameError: (initiate only)
+            BadMechanismError
         """
 
         if self.usage == 'accept':
@@ -406,7 +522,12 @@ class SecurityContext(rsec_contexts.SecurityContext):
         res = rsec_contexts.accept_sec_context(token, self._creds,
                                                self, self._channel_bindings)
 
-        self.delegated_creds = Credentials(res.delegated_creds)
+        if res.delegated_creds is not None:
+            self._delegated_creds = Credentials(res.delegated_creds)
+        else:
+            self._delegated_creds = None
+
+        self._complete = not res.more_steps
 
         return res.token
 
@@ -418,6 +539,8 @@ class SecurityContext(rsec_contexts.SecurityContext):
                                              self._channel_bindings,
                                              token)
 
+        self._complete = not res.more_steps
+
         return res.token
 
     # pickle protocol support
